from .globals import GRID_OUTPUTS, RUNS_DIR, FILE_STORAGE_ROOT
from pyfzf.pyfzf import FzfPrompt
from collections import defaultdict
import incense
import os
import json
from functools import partial
from pyrsistent import thaw
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from tqdm import tqdm
from pathlib import Path
from prettytable import PrettyTable
import pandas as pd

FILESTORAGE_SPECIAL_DIRS = ["_sources", "_resources"]


# %%
def get_incense_loader(authfile="sorcerun_auth.json"):
    with open(authfile, "r") as f:
        js = json.loads(f.read())
    db_name = js["db_name"]
    ck = js["client_kwargs"]
    mongo_uri = f"mongodb://{ck['username']}:{ck['password']}@{ck['host']}:{ck['port']}/{db_name}\?authSource=admin"
    return incense.ExperimentLoader(mongo_uri=mongo_uri, db_name=db_name)


def load_filesystem_expts_by_config_keys(
    runs_dir=f"{FILE_STORAGE_ROOT}/{RUNS_DIR}",
    **kwargs,
):
    runs_dir = Path(runs_dir)
    configs = {
        run_dir.name: incense.experiment._load_json_from_path(run_dir / "config.json")
        for run_dir in runs_dir.iterdir()
        if run_dir.name not in FILESTORAGE_SPECIAL_DIRS
    }

    # kwargs = {"grid_id": "2024-02-27-19-34-08"}

    ids = list(
        filter(
            partial(
                filter_by_dict,
                obj_to_dict=lambda i: squish_dict(thaw(configs[i])),
                **kwargs,
            ),
            configs.keys(),
        )
    )

    expts = [
        incense.experiment.FileSystemExperiment.from_run_dir(runs_dir / i) for i in ids
    ]

    return expts


def filter_by_dict(obj, obj_to_dict=lambda e: squish_dict(thaw(e.config)), **kwargs):
    d = obj_to_dict(obj)
    out = True
    for k, v in kwargs.items():
        if d.get(k, None) == None:
            return False
        out = out and d[k] == v
    return out


def filter_by_config(exps, **kwargs):
    return filter(partial(filter_by_dict, **kwargs), exps)


# %%
def find_differing_keys(ds):
    # find keys that have differing values in the list of dictionaries
    # if a value is a list, convert to tuple

    ranges = defaultdict(set)
    for d in ds:
        for k, v in d.items():
            if type(v) == list:
                v = tuple(v)
            ranges[k].add(v)

    ks = [k for k in ranges if len(ranges[k]) > 1]

    # setted = {k: set(d[k] for d in ds) for k in ds[0]}
    # ks = [k for k in setted if len(setted[k]) > 1]
    return ks


# %%
def squish_dict(d):
    """Flatten a nested dictionary into a single dictionary.

    Nested keys are indexed with a dot. For example,

    d = {
        'a': 1,
        'b': {
            'c': 2,
            'd': 3,
            'blah': {
                'foo': 67,
                'bar': 'no'
            }
        },
        'e': 'lol'
    }

    becomes

    {
        'a': 1,
        'b.c': 2,
        'b.d': 3,
        'b.blah.foo': 67,
        'b.blah.bar': 'no',
        'e': 'lol'
    }

    """
    ks = list(d.keys())
    for k in ks:
        v = d[k]
        assert type(k) == str, "cannot squish non-string keys"
        if type(v) != dict:
            continue

        squished = squish_dict(v)
        for sk, sv in squished.items():
            d[k + "." + sk] = sv

        del d[k]
    return d


# %%
def unsquish_dict(d):
    """Inflate a flattened dictionary back into a nested dictionary.

    Nested keys are indexed with a dot. For example,

    d = {
        'a': 1,
        'b.c': 2,
        'b.d': 3,
        'b.blah.foo': 67,
        'b.blah.bar': 'no',
        'e': 'lol'
    }

    becomes

    {
        'a': 1,
        'b': {
            'c': 2,
            'd': 3,
            'blah': {
                'foo': 67,
                'bar': 'no'
            }
        },
        'e': 'lol'
    }

    """
    result = {}
    for k, v in d.items():
        assert type(k) == str, "cannot unsquish non-string keys"
        keys = k.split(".")
        assert len(keys) > 0, "cannot unsquish keys with no content"

        # Start from the top-level dictionary and gradually drill down
        current_level = result
        for sub_key in keys[:-1]:
            if sub_key not in current_level:
                current_level[sub_key] = {}
            current_level = current_level[sub_key]
            assert isinstance(
                current_level, dict
            ), "non-dict intermediate level detected"

        # By this point, current_level is the dictionary where the value should be placed
        final_key = keys[-1]
        current_level[final_key] = v
    return result


# %%
def exps_to_xarray(exps, exclude_keys=["seed"]):
    """Convert a list of incense experiments to two xarrays.

    :param exps: input list of experiments
    :param exclude_keys: config keys to exclude

    Returns: tuple (exps_arr, metrics_arr)
    Both are xarrays.

    exps_arr has dims given by the union of squished config keys over all experiments.
    The corresponding coords are also the union over all possible values of the
    corresponding config key. The value of each entry is the incense Experiment object.

    metrics_arr is the same as exps_arr, with two additional dims: "metric" and "step".
    The value of each entry is now the value of the metric at step "step" for each
    experiment.
    """
    # info about tuple/list valued coordinates:
    # First of all, when extracting the coordinates, we need to convert list values to tuples
    # because lists are unhashable and we use set unioning to get the unique values of each
    # coordinate.
    # Second, if you pass a list of tuples as coordinate values to xarray, it doesnt treat
    # it as a 1D coordinate, but as a 2D coordinate. This is not what we want. So we convert
    # the list of tuples to a numpy array of objects, which xarray treats as a 1D coordinate.

    # Finally, when saving this to netcdf, xarray didnt like tuple values as coordinates
    # either, so we just convert them to strings, only for saving to netcdf.
    # This change is not in this function, but in process_and_save_grid_to_netcdf.

    # get set of axis keys from config
    e_cfg_keys = set()
    for e in exps:
        e_cfg_keys = e_cfg_keys.union(set(squish_dict(thaw(e.config)).keys()))
    axes_without_metric = list(e_cfg_keys - set(exclude_keys))
    axes = axes_without_metric + ["metric", "step"]
    print(f"axes={axes}")

    e = exps[1]
    # get coordinates of each axis key in the grid
    # by looping through every experiment and taking the union of the coord set
    coords = {a: set() for a in axes}
    for e in tqdm(exps, desc="Extracting coords"):
        e_cfg = squish_dict(thaw(e.config))
        for a in axes:
            if a == "metric":
                # union all the metrics
                s = set(e.metrics.keys())
            elif a == "step":
                # union over the possible steps for all metrics for this exp
                s = set()
                for v in e.metrics.values():
                    s = s.union(list(v.index))
            else:
                val = e_cfg[a]
                # both list and np.array are not hashable, so convert to tuple
                if type(val) == list:
                    val = tuple(val)
                s = set([val])

            # union over all the experiments
            coords[a] = coords[a].union(s)
    for k, v in coords.items():
        if any(type(x) == tuple for x in v):
            vl = list(v)
            # now use np.array because xarray doesn't like lists of tuples
            vals = np.empty(shape=(len(vl),), dtype=object)
            for i, val in enumerate(vl):
                vals[i] = val
        else:
            vals = sorted(list(v))
        coords[k] = vals

    # remove axes of size 1 (except metric and step)

    axes_copy = axes.copy()
    size_one_axes = []
    for a in axes_copy:
        if a == "metric" or a == "step":
            continue
        if len(coords[a]) == 1:
            print(f"Removing axis {a} because it has size 1")
            axes.remove(a)
            axes_without_metric.remove(a)
            coords.pop(a)
            size_one_axes.append(a)

    # print out the axes and their sizes
    t = PrettyTable(["Axis", "Size"])
    for a in axes:
        t.add_row([a, len(coords[a])])
    t.align = "l"
    print(t)

    coords_without_metric = coords.copy()
    coords_without_metric.pop("metric")
    coords_without_metric.pop("step")

    # create empty xarrays
    shape = tuple(len(coords[a]) for a in axes)

    # metric data should have same dtype as the metric values
    metric_dtype = np.float64
    metric_data = np.empty(shape, dtype=metric_dtype)
    metric_data.fill(np.nan)
    metrics_arr = xr.DataArray(
        metric_data,
        coords=coords,
        dims=axes,
        name="metrics",
    )

    shape_without_metric = tuple(len(coords[a]) for a in axes_without_metric)
    exps_data = np.empty(shape_without_metric, dtype=object)
    exps_data.fill(np.nan)
    exps_arr = xr.DataArray(
        exps_data,
        coords=coords_without_metric,
        dims=axes_without_metric,
        name="experiments",
    )

    # now fill in the metric values
    for e in tqdm(exps[::-1], desc="Filling in metrics"):
        e_cfg_index = squish_dict(thaw(e.config))
        [e_cfg_index.pop(k) for k in exclude_keys]
        [e_cfg_index.pop(k) for k in size_one_axes]

        # convert values of type list to tuple
        for k, v in e_cfg_index.items():
            if type(v) == list:
                e_cfg_index[k] = tuple(v)

        exps_arr.loc[e_cfg_index] = e

        # get the x_arr of this experiment
        e_arr = metrics_arr.sel(**e_cfg_index)

        for k, v in e.metrics.items():
            # fill in each metric into the x_arr
            # Note that we are only filling in values of v.index, and this will
            # always be a subset of the coordspace of step since we did the unioning
            # before.
            e_arr.loc[dict(metric=k, step=v.index)] = v.values

    return exps_arr, metrics_arr


def print_file_size(file_path):
    file_info = os.stat(file_path)
    size_in_bytes = file_info.st_size

    if size_in_bytes < 1024:
        print(f"The size of '{file_path}' is: {size_in_bytes} bytes")
    elif size_in_bytes >= 1024 and size_in_bytes < 1024**2:
        print(f"The size of '{file_path}' is: {size_in_bytes / 1024:.2f} KB")
    elif size_in_bytes >= 1024**2 and size_in_bytes < 1024**3:
        print(f"The size of '{file_path}' is: {size_in_bytes / 1024**2:.2f} MB")
    else:
        print(f"The size of '{file_path}' is: {size_in_bytes / 1024**3:.2f} GB")


def process_and_save_grid_to_netcdf(gid, file_root=FILE_STORAGE_ROOT):
    grid_exps = load_filesystem_expts_by_config_keys(
        grid_id=gid,
        runs_dir=os.path.join(file_root, RUNS_DIR),
    )

    print(f"Found {len(grid_exps)} experiments with grid_id {gid}")

    grid_exps_xr, grid_metrics_xr = exps_to_xarray(grid_exps)
    metrics_reduced_xr = grid_metrics_xr

    save_dir = f"{file_root}/{GRID_OUTPUTS}/{gid}"
    os.makedirs(save_dir, exist_ok=True)

    netcdf_save_path = f"{save_dir}/{gid}.nc"
    print(f"Saving to {netcdf_save_path}")

    # convert coordinate values that are tuples to strings to avoid serialization issues
    old_coords = [(k, v) for k, v in metrics_reduced_xr.coords.items()]
    for k, v in old_coords:
        if any(type(x) == tuple for x in v.values):
            print(f"Converting coordinate values of {k} to str")
            metrics_reduced_xr.coords[k] = [str(x) for x in v.values]

    metrics_reduced_xr.to_netcdf(netcdf_save_path)

    print_file_size(netcdf_save_path)


# %%
def xarray_to_csv(da, csv_filename):
    """
    Convert an xarray DataArray to a CSV file, removing NaN values.

    Parameters:
    da (xarray.DataArray): The input DataArray
    csv_filename (str): The name of the output CSV file

    Returns:
    None
    """
    # Step 1: Convert to pandas DataFrame
    df = da.to_dataframe()

    # Step 2: Drop NaN values
    df_clean = df.dropna()

    # Step 3: Reset index to create columns for each coordinate
    df_clean_reset = df_clean.reset_index()

    # Step 4: Save to CSV
    df_clean_reset.to_csv(csv_filename, index=False)

    print_file_size(csv_filename)

    return df_clean_reset


# %%
def csv_to_xarray(csv_filename, value_column):
    """
    Convert a CSV file to an xarray DataArray,
    treating all columns except the value column as coordinates.
    The DataArray is named after the value column.

    Parameters:
    csv_filename (str): The name of the input CSV file
    value_column (str): The name of the column containing the data values

    Returns:
    xarray.DataArray: The resulting DataArray
    """
    df = pd.read_csv(csv_filename)

    coord_columns = [col for col in df.columns if col != value_column]

    coords = {col: sorted(df[col].unique()) for col in coord_columns}

    shape = tuple(len(coords[col]) for col in coord_columns)
    data = np.full(shape, np.nan)

    for _, row in df.iterrows():
        index = tuple(coords[col].index(row[col]) for col in coord_columns)
        data[index] = row[value_column]

    da = xr.DataArray(data, coords=coords, dims=coord_columns, name=value_column)

    return da


# %%
def process_and_save_grid_to_csv(gid, file_root=FILE_STORAGE_ROOT):
    grid_exps = load_filesystem_expts_by_config_keys(
        grid_id=gid,
        runs_dir=os.path.join(file_root, RUNS_DIR),
    )

    print(f"Found {len(grid_exps)} experiments with grid_id {gid}")

    grid_exps_xr, grid_metrics_xr = exps_to_xarray(grid_exps)

    save_dir = f"{file_root}/{GRID_OUTPUTS}/{gid}"
    os.makedirs(save_dir, exist_ok=True)

    csv_filename = f"{save_dir}/{gid}.csv"
    print(f"Saving to {csv_filename}")

    # convert coordinate values that are tuples to strings to avoid serialization issues
    old_coords = [(k, v) for k, v in grid_metrics_xr.coords.items()]
    for k, v in old_coords:
        if any(type(x) == tuple for x in v.values):
            print(f"Converting coordinate values of {k} to str")
            grid_metrics_xr.coords[k] = [str(x) for x in v.values]

    df = xarray_to_csv(grid_metrics_xr, csv_filename)
    return df


# %%
def get_latest_single_and_grid_exps(exps):
    """Parse list of experiments to get the latest
    single experiment and grid of experiments
    (a grid of exps is assumed to have the
    same time_str for all exps in the grid).

    :param exps: list of incense experiments

    Returns: tuple (single_exp, grid_exps)

    If no single experiment is found, single_exp is None.
    If no grid experiments are found, grid_exps is an empty list.
    """
    time_str_to_exps = {}
    for exp in exps:
        time_str = exp.config.time_str
        if time_str not in time_str_to_exps:
            time_str_to_exps[time_str] = []
        time_str_to_exps[time_str].append(exp)
    sorted_time_strs = sorted(time_str_to_exps.keys())[::-1]
    single_exp_time_str = None
    grid_time_str = None
    for time_str in sorted_time_strs:
        if len(time_str_to_exps[time_str]) == 1:
            single_exp_time_str = (
                time_str if single_exp_time_str is None else single_exp_time_str
            )
        else:
            grid_time_str = time_str if grid_time_str is None else grid_time_str

        if single_exp_time_str is not None and grid_time_str is not None:
            break

    single_exp = time_str_to_exps.get(single_exp_time_str, [None])[0]
    grid_exps = time_str_to_exps.get(grid_time_str, [])
    return single_exp, grid_exps


# %%
def dict_to_fzf_friendly_str(d):
    return " ".join([f"{k}:{v}" for k, v in d.items()]) + " "


# this class is still WIP
class ExpsSlicer:
    def __init__(self, exps, runs_dir=RUNS_DIR, exclude_keys=["seed"]):
        self.exps = exps
        self.runs_dir = runs_dir
        self.dicts = [squish_dict(thaw(e.config)) for e in self.exps]
        for k in exclude_keys:
            [d.pop(k) for d in self.dicts]

        self.ks = find_differing_keys(self.dicts)
        dicts_with_only_diff_keys = [
            {k: d.get(k, None) for k in self.ks} for d in self.dicts
        ]
        ids = [str(e.id) for e in self.exps]
        self.id_to_exp = {i: e for i, e in zip(ids, self.exps)}
        self.dictstrs = [
            f"{i} " + dict_to_fzf_friendly_str(d)
            for i, d in zip(ids, dicts_with_only_diff_keys)
        ]

        self.ranges = defaultdict(set)
        for d in dicts_with_only_diff_keys:
            for k, v in d.items():
                self.ranges[k].add(v)

        self.ranges = {k: sorted(list(v)) for k, v in self.ranges.items()}
        self.ranges_dir = os.path.join(FILE_STORAGE_ROOT, "ranges")
        os.makedirs(self.ranges_dir, exist_ok=True)
        for k, v in self.ranges.items():
            with open(f"{self.ranges_dir}/{k}_range.txt", "w") as f:
                f.write("\n".join(map(str, v)))

        self.fzf = FzfPrompt()

    def fzf_filter(self):
        key = self.fzf.prompt(
            list(self.ks), " --preview 'cat " + f"{self.ranges_dir}/" + "{1}_range.txt'"
        )[0]
        out = self.fzf.prompt(
            self.dictstrs,
            "--multi"
            + " --preview 'cat "
            + f"{self.runs_dir}/"
            + "{1}/cout.txt'"
            + f" --print-query"
            + f" --query \\'{key}:"
            + " --bind enter:select-all+accept",
        )
        query = out[0]
        out_ids = [o.split()[0] for o in out[1:]]
        out_exps = [self.id_to_exp[i] for i in out_ids]
        return ExpsSlicer(out_exps, runs_dir=self.runs_dir)

    def __call__(self, **kwargs):
        if len(kwargs) == 0:
            return self.fzf_filter()
        return ExpsSlicer(
            list(filter_by_config(self.exps, **kwargs)),
            runs_dir=self.runs_dir,
        )

    def __len__(self):
        return len(self.exps)


# %%
if __name__ == "__main__":
    fzf = FzfPrompt()
    gid = fzf.prompt(os.listdir(f"{FILE_STORAGE_ROOT}/{GRID_OUTPUTS}"))[0]
    runs_dir = os.path.join(FILE_STORAGE_ROOT, RUNS_DIR)
    out = load_filesystem_expts_by_config_keys(
        grid_id=gid,
        runs_dir=runs_dir,
    )
    print(len(out))
    exps_arr, metrics_arr = exps_to_xarray(out, exclude_keys=["seed"])

    # product of all the axes sizes
    print(np.prod([len(v) for v in metrics_arr.coords.values()]))

    df = xarray_to_csv(metrics_arr, f"{gid}.csv")
    da = csv_to_xarray(f"{gid}.csv", "metrics")
    err = np.abs(da.data - metrics_arr.data)
    print(np.sum(err[~np.isnan(err)]))

#     slicer = ExpsSlicer(out, runs_dir=runs_dir)
#     e = out[0]
#     print(e._data["captured_out"])
#     vars(e)
#     ans = slicer()()()



import json
from sacred.commandline_options import CommandLineOption
from sacred.observers import MongoObserver
from pymongo import MongoClient


class AuthMongoDbOption(CommandLineOption):
    """Custom MongoObserver option with authentication.
    The authentication json file should contain two keys:
        'client_kwargs': directly passed to MongoClient
        'db_name': name of database to store runs (usually sacred)"""

    arg = "AUTH_FILE"
    short_flag = "M"
    arg_description = """Path to authentication json file."""

    @classmethod
    def apply(cls, args, run):
        with open(args) as f:
            auth = json.loads(f.read())
        atlas = auth.get("atlas_connection_string", 0)
        if atlas != 0:
            print(atlas)
            client = MongoClient(atlas)
        else:
            client = MongoClient(**auth["client_kwargs"])
        mongo = MongoObserver.create(db_name=auth["db_name"], client=client)
        run.observers.append(mongo)



from numpy import squeeze
import numpy as np
from sorcerun.globals import GRID_OUTPUTS
import itertools
import os
import xarray as xr
import matplotlib.pyplot as plt
import streamlit as st
import scipy.stats
from sorcerun.incense_utils import csv_to_xarray

st.title("Grid Plotter")
# show current working directory
working_dir = os.getcwd()
st.write(f"Current working directory: `{working_dir}`")

st.title("Settings")

# Checkbox to decide whether to show all grid ids or only the ones with tags
show_all_grids = not st.checkbox("Show only tagged grids", value=False)

# Extract grid ids
name_to_gid = {}
for f in os.listdir(GRID_OUTPUTS):
    if os.path.isdir(os.path.join(GRID_OUTPUTS, f)):
        gid = f
        tags_file = os.path.join(GRID_OUTPUTS, gid, f"{gid}-tags.txt")

        tags = ""
        if os.path.exists(tags_file):
            with open(tags_file, "r") as f:
                tags = f" tags: {f.read().strip()}"

        if show_all_grids or tags:
            name_to_gid[f"{gid}{tags}"] = gid


# Select a grid id
name = st.selectbox("Choose a Grid ID", list(name_to_gid.keys()))
grid_id = name_to_gid[name]

# Add a text box to modify tags for the selected grid id
tags_file = os.path.join(GRID_OUTPUTS, grid_id, f"{grid_id}-tags.txt")
tags = ""
if os.path.exists(tags_file):
    with open(tags_file, "r") as f:
        tags = f.read().strip()

new_tags = st.text_input("Tags", value=tags)

# Add a button to save the tags
if st.button("Save tags"):
    # delete the tags file if the new tags are empty
    if not new_tags:
        if os.path.exists(tags_file):
            os.remove(tags_file)
    else:
        with open(tags_file, "w") as f:
            f.write(new_tags)
    st.rerun()

st.write(f"Selected grid id: `{grid_id}`")


# Load the csv file for the selected grid id as an xarray
grid_output_dir = os.path.join(GRID_OUTPUTS, grid_id)
csv_file = os.path.join(grid_output_dir, f"{grid_id}.csv")

data = csv_to_xarray(csv_file, value_column="metrics")

# get dims with more than 1 coordinate
dims = sorted([dim for dim in data.dims if len(data[dim]) > 1 and dim != "metric"])

# get metric names

# For each axis, have an option to reduce the coordinate by taking mean, max, min or just ignoring it
# Use a radio button to select the reduction option
reduce_options = [
    "don't reduce",
    "mean",
    "max",
    "min",
]


# create a list of strings showing shorted coordinate values
# (if the length of the coordinate is greater than 10, then it is shortened)
MAX_LEN = 100
shortened_coords = {
    dim: (
        str(data.coords[dim].values)
        if len(str(data.coords[dim].values)) < MAX_LEN
        else str(data.coords[dim].values)[: MAX_LEN // 2]
        + "..."
        + str(data.coords[dim].values)[-MAX_LEN // 2 :]
    )
    for dim in dims
}

st.subheader("Dimensions with more than 1 coordinate value")
st.write("Select a reduction option for each dimension")
# create a dictionary of reduction options for each dimension
# if the dim is `repeat`, then the default reduction option is `mean` by default
reduce_options_dict = {
    dim: (
        st.radio(
            f"`{dim}`: {shortened_coords[dim]}",
            reduce_options,
            horizontal=True,
        )
        if dim != "repeat"
        else st.radio(
            f"`{dim}`: {shortened_coords[dim]}",
            reduce_options,
            horizontal=True,
            index=1,
        )
    )
    for dim in dims
}

# apply the reduction to the data
for dim, reduce_option in reduce_options_dict.items():
    if reduce_option == "mean":
        data = data.mean(dim=dim)
    elif reduce_option == "max":
        data = data.max(dim=dim)
    elif reduce_option == "min":
        data = data.min(dim=dim)

# Get the new set of axes
new_dims = [dim for dim in data.dims if len(data[dim]) > 1]

metric_names = list(data["metric"].values)
st.subheader("Metrics")
st.write(metric_names)

st.subheader("Choose axes for each plot")
xlim = None
ylim = None
# Choose an x and y axis from both dimensions and metrics
with st.container():
    # add a column for setting xlim and ylim
    col1x, col2x, col3x, col4x = st.columns(4)
    with col1x:
        x_axis = st.selectbox(
            "X axis",
            new_dims,
            index=new_dims.index("step") if "step" in new_dims else 0,
        )
    with col2x:
        st.text("")
        st.text("")
        log_x = st.checkbox("Log scale x axis", value=False)

    with col3x:
        x_label = st.text_input("X label", value=x_axis)

    with col4x:
        xlim = st.text_input("X limits", value="")
        if xlim:
            xlim = [float(x) for x in xlim.split(",")]

    col1y, col2y, col3y, col4y = st.columns(4)
    # Choose multiple y axes from the metrics
    with col1y:
        y_axes = st.multiselect("Y axis", metric_names)
    with col2y:
        st.text("")
        st.text("")
        log_y = st.checkbox("Log scale y axis", value=False)
    with col3y:
        y_label = st.text_input("Y label", value="metric")
    with col4y:
        ylim = st.text_input("Y limits", value="")
        if ylim:
            ylim = [float(y) for y in ylim.split(",")]

# Remaining dims
remaining_dims = sorted(list(set(new_dims) - set([x_axis, "metric"])))

# Choose a style for the plot
style = st.text_input("Line style", value="o-")

# Choose plot dimensions to include within a single plot
st.subheader("Choose dimensions that are kept fixed in each plot")
fixed_dims_per_plot = st.multiselect(
    "Fixed dims per plot",
    remaining_dims,
)

dims_used_in_plot = sorted(list(set(remaining_dims) - set(fixed_dims_per_plot)))
other_dims = fixed_dims_per_plot
st.write("Dimensions that vary in each plot:")
st.write(dims_used_in_plot)

# Boolean on whether to show regression line slope
show_slope = st.checkbox("Show regression line slope", value=False)

# Boolean on whether to have axis grids
axis_grids = st.checkbox("Show axis grids", value=True)

# Boolean on whether to save the plots
save_plots = st.checkbox("Save plots", value=False)

if save_plots:
    # Select a save type (png, eps, pdf)
    save_type = st.selectbox("Save type", ["png", "eps", "pdf"])
    save_dir = f"figures/{grid_id}/{save_type}"
    st.write(f"Saving plots to: `{save_dir}`")
    os.makedirs(save_dir, exist_ok=True)

# ToDO: auto change ylabel is there is only one metric and dont include that in legend

# Add a button to actually make the plots
# check if other_dims is empty
if not other_dims:
    out = data
    groups = [()]
else:
    out = data.stack(other_dims=other_dims).groupby("other_dims", squeeze=False)
    groups = list(out.groups.keys())

if st.button(f"Make {len(groups)} plots"):
    st.subheader(f"Creating {len(groups)} plots")
    st.write(f"Plotting `{x_axis}` vs `{y_axes}` for each group of `{other_dims}`")

    TITLE_NAME = {k: k for k in other_dims}
    title_keys = other_dims

    x = data.coords[x_axis].values

    for i, group in enumerate(groups):
        title = "--".join(
            [
                f"{TITLE_NAME[key]}={val}"
                for key, val in zip(other_dims, group)
                if key in title_keys
            ]
            + [f"metrics={'-'.join(y_axes)}"]
        )
        xarr = out[group]

        fig, ax = plt.subplots()
        plt.title(title)
        # use itertools product to iterate over all coordinate combinations of dims_used_in_plot
        for coord in itertools.product(
            *[xarr.coords[dim].values for dim in dims_used_in_plot]
        ):
            d = {dim: coord[i] for i, dim in enumerate(dims_used_in_plot)}
            ys = xarr.loc[d]
            for y_axis in y_axes:
                y = ys.loc[dict(metric=y_axis)].squeeze().to_numpy()
                label = ", ".join([f"{k}={v}" for k, v in d.items()])
                if len(y_axes) > 1:
                    label = y_axis + " " + label

                # check if y is all nans or infs, if so, don't plot
                good_inds = np.isfinite(y)

                # dont_plot = y.isnull().all().item() or (y == float("inf")).all().item()
                dont_plot = not np.any(good_inds)
                if not dont_plot:
                    plt.plot(x, y, style, label=label)
                    # add regression line slope
                    if len(x) > 1 and show_slope:
                        x_to_regress = np.log(x) if log_x else x
                        y_to_regress = np.log(y) if log_y else y

                        # remove nans and infs
                        keep_inds = np.isfinite(x_to_regress) & np.isfinite(
                            y_to_regress
                        )
                        x_to_regress = x_to_regress[np.where(keep_inds)]
                        y_to_regress = y_to_regress[np.where(keep_inds)]

                        # slope, intercept = np.polyfit(x_to_regress, y_to_regress, 1)
                        (
                            slope,
                            intercept,
                            r_value,
                            p_value,
                            std_err,
                        ) = scipy.stats.linregress(x_to_regress, y_to_regress)
                        plt.text(
                            x[-1],
                            y[-1],
                            f"Slope: {slope:.7f}, $r^2$: {r_value**2:.7f}",
                            horizontalalignment="right",
                            verticalalignment="top",
                        )

        if log_x:
            plt.xscale("log")
        if log_y:
            plt.yscale("log")
        plt.ylabel(y_label)
        plt.xlabel(x_label)
        if xlim:
            plt.xlim(xlim)
        if ylim:
            plt.ylim(ylim)

        # place legend outside plot to the right using axis
        leg = fig.legend(loc="center left", bbox_to_anchor=(0.92, 0.5))
        # plt.legend()

        if axis_grids:
            plt.grid(True)

        # display plot in streamlit
        if save_plots:
            fig.savefig(
                f"{save_dir}/{i:03d}--{title}.{save_type}",
                bbox_extra_artists=[leg],
                bbox_inches="tight",
            )
        st.write(f"Plot {i+1}: {title}")
        st.pyplot(fig)



import os
import yaml, json
import contextlib
import subprocess
import click
from pymongo import MongoClient
from .globals import AUTH_FILE, MONGOD_PORT, MONGOD_HOST


@contextlib.contextmanager
def mongodb_server(conf_path):
    click.echo(f"Starting MongoD server using config at {conf_path}")
    mongodb_process = subprocess.Popen(
        ["mongod", "--config", conf_path],
        stdout=subprocess.DEVNULL,  # Suppress output
    )

    # read from log file using tail and then clean it up using jq
    with open(conf_path, "r") as conf_file:
        conf = yaml.load(conf_file.read(), Loader=yaml.Loader)
        log_path = conf["systemLog"]["path"]

    # Failed attempt at reading important stuff from logs and printing on a line

    # separator = ", "
    # jq_command = (
    #     "["
    #     + separator.join(
    #         [
    #             '.t."$date"',
    #             r".s",
    #             r".c",
    #             r".id",
    #             r".ctx",
    #             r".msg",
    #             r".attr.mechanism",
    #             r".attr.principalName",
    #             r".attr.remote",
    #         ]
    #     )
    #     + "]"
    # )

    jq_command = "."
    taillog_process = subprocess.Popen(
        f"tail -f {log_path} | jq '{jq_command}'",
        shell=True,
    )
    try:
        yield
    finally:
        click.echo(f"Terimnating MongoD server that used config at {conf_path}\n")
        mongodb_process.terminate()
        mongodb_process.wait()

        taillog_process.terminate()
        taillog_process.wait()


def create_mongod_conf(db_path, conf_path, auth=True):
    conf_data = {
        "storage": {
            "dbPath": db_path,
        },
        "net": {
            "bindIp": MONGOD_HOST,
            "port": MONGOD_PORT,
        },
        "security": {
            "authorization": "enabled" if auth else "disabled",
        },
        "systemLog": {
            "destination": "file",
            "path": os.path.join(db_path, "mongod.log"),
            "logAppend": True,
        },
    }

    with open(conf_path, "w") as conf_file:
        yaml.dump(conf_data, conf_file, default_flow_style=False)


def init_mongodb(db_path, db_name, username, password):
    db_path = os.path.abspath(db_path)
    if not os.path.exists(db_path):
        os.makedirs(db_path)

    # --- create admin user with no auth ---
    conf_path_noauth = os.path.join(db_path, "mongod_noauth.conf")
    create_mongod_conf(db_path, conf_path_noauth, auth=False)

    with mongodb_server(conf_path_noauth):
        click.echo("Creating admin user in mongodb")
        client = MongoClient(MONGOD_HOST, MONGOD_PORT)
        db = client["admin"]
        db.command("ping")
        client.admin.command(
            "createUser",
            "admin",
            pwd="admin",
            roles=[{"role": "userAdminAnyDatabase", "db": "admin"}],
        )

    # remove the conf_file with no auth

    # --- create specific user with auth for database ---
    conf_path = os.path.join(db_path, "mongod.conf")
    create_mongod_conf(db_path, conf_path)

    with mongodb_server(conf_path):
        click.echo(f"Creating user {username} with access to database {db_name}")
        # client authed with as admin
        client = MongoClient(
            f"mongodb://admin:admin@{MONGOD_HOST}:{MONGOD_PORT}/?authSource=admin"
        )
        db = client[db_name]
        db.command("ping")

        client.admin.command(
            "createUser",
            username,
            pwd=password,
            roles=[{"role": "readWrite", "db": db_name}],
        )

        # 8. Create auth.json file
        auth_data = {
            "client_kwargs": {
                "host": MONGOD_HOST,
                "port": MONGOD_PORT,
                "username": username,
                "password": password,
                "authSource": "admin",
            },
            "db_name": db_name,
            "db_path": db_path,
            "conf_path": conf_path,
        }

        with open(AUTH_FILE, "w") as auth_file:
            json.dump(auth_data, auth_file, indent=4)



from sorcerun.git_utils import (
    is_dirty,
    get_repo,
    get_commit_hash,
    get_time_str,
    get_tree_hash,
)

repo = get_repo()
commit_hash = get_commit_hash(repo)
time_str = get_time_str()
main_tree_hash = (get_tree_hash(repo, "main"),)
dirty = is_dirty(repo)
grid_id = (f"{time_str}--{commit_hash}--dirty={dirty}",)

ns = [10, 20, 30]


def make_config(n):
    c = {
        "n": n,
        "commit_hash": commit_hash,
        "main_tree_hash": main_tree_hash,
        "time_str": time_str,
        "dirty": dirty,
        "grid_id": grid_id,
    }
    return c


configs = [make_config(n) for n in ns]



from sorcerun.sacred_utils import run_sacred_experiment


def adapter(config, _run):
    n = config["n"]

    for i in range(n):
        _run.log_scalar("i", i)
        _run.log_scalar("i_squared", i**2)
        _run.log_scalar("i_cubed", i**3)

    return 0


adapter.experiment_name = "sample_experiment"

if __name__ == "__main__":
    from config import config

    run_sacred_experiment(adapter, config)



from sorcerun.git_utils import (
    is_dirty,
    get_repo,
    get_commit_hash,
    get_time_str,
    get_tree_hash,
)

repo = get_repo()

n = 10

config = {
    "n": n,
    "commit_hash": get_commit_hash(repo),
    "main_tree_hash": get_tree_hash(repo, "main"),
    "time_str": get_time_str(),
    "dirty": is_dirty(get_repo()),
}



import click
import time
from sklearn.model_selection import ParameterGrid
import os
import subprocess
import json, yaml
from contextlib import ExitStack
from .mongodb_utils import mongodb_server, init_mongodb
from .sacred_utils import load_python_module, run_sacred_experiment
from .incense_utils import (
    squish_dict,
    unsquish_dict,
    process_and_save_grid_to_netcdf,
    process_and_save_grid_to_csv,
)
from .globals import (
    AUTH_FILE,
    TEMP_CONFIGS_DIR,
    FILE_STORAGE_ROOT,
    GRID_OUTPUTS,
    TEMPLATE_FILES,
)
from .slurm_utils import Job, poll_jobs

from prettytable import PrettyTable
import sys, ipdb, traceback


def info(type, value, tb):
    traceback.print_exception(type, value, tb)
    ipdb.pm()


@click.group()
@click.option(
    "--debug",
    "-d",
    is_flag=True,
    help="debug",
)
def sorcerun(debug):
    if debug:
        sys.excepthook = info


@sorcerun.command()
def template():
    # TODO: update this to use subtrees and generate notebooks with some plots instead of the streamlit app
    this_file_path = os.path.abspath(__file__)
    this_dir = os.path.dirname(this_file_path)

    for file in TEMPLATE_FILES:
        # check if file already exists in current directory
        if os.path.exists(file):
            click.echo(f"File {file} already exists. Skipping creation.")
            continue

        path_to_template_file = os.path.join(this_dir, file)
        with open(path_to_template_file, "r") as f:
            click.echo(f"Creating file: {file}")
            with open(file, "w") as new_f:
                new_f.write(f.read())


@sorcerun.command()
@click.argument("python_file", type=click.Path(exists=True, dir_okay=False))
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "--file_root",
    "-f",
    default=FILE_STORAGE_ROOT,
    type=click.Path(file_okay=False),
    help="Root directory for file storage",
)
@click.option("--auth_path", default=AUTH_FILE, help="Path to sorcerun_auth.json file.")
@click.option(
    "--mongo",
    "-m",
    is_flag=True,
    help="Use MongoObserver",
)
@click.option(
    "--dont_profile",
    is_flag=True,
    help="Don't use cProfile to profile the adapter function",
    default=False,
)
def run(
    python_file,
    config_file,
    file_root,
    auth_path,
    mongo,
    dont_profile,
):
    # Load the adapter function from the provided Python file
    adapter_module = load_python_module(python_file)
    if not hasattr(adapter_module, "adapter"):
        raise KeyError(
            f"Adapter file at {python_file} does not have an attribute named adapter"
        )
    adapter_func = adapter_module.adapter

    _, config_ext = os.path.splitext(config_file)

    # Check extension of config file and load it accordingly
    if config_ext == ".json":
        with open(config_file, "r") as file:
            config = json.load(file)
    elif config_ext == ".yaml":
        with open(config_file, "r") as file:
            config = yaml.safe_load(file)
    elif config_ext == ".py":
        config_module = load_python_module(config_file)
        if not hasattr(config_module, "config"):
            raise KeyError(
                f"Config file at {config_file} does not have an attribute named config"
            )
        config = config_module.config
    else:
        raise ValueError(
            f"Config file at {config_file} is not a valid JSON, YAML or python file"
        )

    # Run the Sacred experiment with the provided adapter function and config
    run_sacred_experiment(
        adapter_func,
        config,
        auth_path,
        use_mongo=mongo,
        file_storage_root=file_root,
        profile=not dont_profile,
    )


@sorcerun.command()
@click.argument(
    "python_file",
    type=click.Path(exists=True, dir_okay=False),
)
@click.argument(
    "grid_config_file",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--file_root",
    "-f",
    default=FILE_STORAGE_ROOT,
    type=click.Path(file_okay=False),
    help="Root directory for file storage",
)
@click.option(
    "--auth_path",
    default=AUTH_FILE,
    help="Path to sorcerun_auth.json file.",
)
@click.option(
    "--post_process",
    "-p",
    is_flag=True,
    help="Post process and save grid to netcdf",
)
@click.option(
    "--mongo",
    "-m",
    is_flag=True,
    help="Use MongoObserver",
)
def grid_run(
    python_file,
    grid_config_file,
    file_root,
    auth_path,
    post_process=False,
    mongo=False,
):
    grid_run_function(
        python_file,
        grid_config_file,
        file_root,
        auth_path,
        post_process=post_process,
        mongo=mongo,
    )


def grid_run_function(
    python_file,
    grid_config_file,
    file_root=FILE_STORAGE_ROOT,
    auth_path=AUTH_FILE,
    post_process=False,
    mongo=False,
):
    # Load the adapter function from the provided Python file
    adapter_module = load_python_module(python_file)
    if not hasattr(adapter_module, "adapter"):
        raise KeyError(
            f"Adapter file at {python_file} does not have an attribute named adapter"
        )
    adapter_func = adapter_module.adapter
    pre_grid_hook = getattr(adapter_module, "pre_grid_hook", None)
    post_grid_hook = getattr(adapter_module, "post_grid_hook", None)

    # Check extension of grid config file and load it accordingly
    _, config_ext = os.path.splitext(grid_config_file)

    if config_ext == ".yaml":
        with open(grid_config_file, "r") as file:
            config = yaml.safe_load(file)
        config = squish_dict(config)
        for k, v in config.items():
            if type(v) != list:
                config[k] = [v]
        param_grid = ParameterGrid([config])
        configs = [unsquish_dict(param) for param in param_grid]

    elif config_ext == ".py":
        config_module = load_python_module(grid_config_file)
        if not hasattr(config_module, "configs"):
            raise KeyError(
                f"Config file at {grid_config_file} does not have an attribute named configs"
            )
        configs = config_module.configs
    else:
        raise ValueError(
            f"Config file at {grid_config_file} is not a valid YAML or python file"
        )

    total_num_params = len(configs)
    print(f"Config grid contains {total_num_params} combinations")

    # Run the Sacred experiment with the provided adapter function and config
    for i, conf in enumerate(configs):
        print(
            "-" * 5
            + "GRID RUN INFO: "
            + f"Starting run {i+1}/{total_num_params}"
            + "-" * 5
        )
        if pre_grid_hook is not None:
            print(f"Running pre_grid_hook")
            pre_grid_hook(conf)

        print("Running experiment with config:")
        print(json.dumps(conf, indent=2))

        run_sacred_experiment(
            adapter_func,
            conf,
            auth_path,
            use_mongo=mongo,
            file_storage_root=file_root,
        )

        if post_grid_hook is not None:
            print(f"Running post_grid_hook")
            post_grid_hook(conf)

        print(
            "-" * 5
            + "GRID RUN INFO: "
            + f"Completed run {i+1}/{total_num_params}"
            + "-" * 5
        )

    if post_process:
        # Post process grid and save xarray to netcdf
        # check if each config in configs has the same "grid_id" and assign it to gid
        gid = configs[0].get("grid_id", None)
        same_gid = False
        if gid is not None:
            same_gid = all(conf.get("grid_id", None) == gid for conf in configs)

        if same_gid:
            print(f"All configs have the same grid_id: {gid}")
            print(f"Processing and saving grid to csv")
            process_and_save_grid_to_csv(gid, file_root=file_root)
        else:
            print(
                f"Configs do not have the same grid_id."
                + " Skipping processing and saving grid to csv"
            )


# %%
@sorcerun.command()
@click.argument(
    "python_file",
    type=click.Path(exists=True, dir_okay=False),
)
@click.argument(
    "grid_config_file",
    type=click.Path(exists=True, dir_okay=False),
)
@click.argument(
    "slurm_config_file",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--file_root",
    "-f",
    default=FILE_STORAGE_ROOT,
    type=click.Path(file_okay=False),
    help="Root directory for file storage",
)
@click.option(
    "--auth_path",
    default=AUTH_FILE,
    help="Path to sorcerun_auth.json file.",
)
@click.option(
    "--post_process",
    "-p",
    is_flag=True,
    help="Post process and save grid to netcdf",
)
@click.option(
    "--mongo",
    "-m",
    is_flag=True,
    help="Use MongoObserver",
)
def grid_slurm(
    python_file,
    grid_config_file,
    slurm_config_file,
    file_root,
    auth_path,
    post_process=False,
    mongo=False,
):
    # Load the adapter function from the provided Python file
    adapter_module = load_python_module(python_file)
    if not hasattr(adapter_module, "adapter"):
        raise KeyError(
            f"Adapter file at {python_file} does not have an attribute named adapter"
        )

    # Check extension of grid config file and load it accordingly
    _, config_ext = os.path.splitext(grid_config_file)

    if config_ext == ".yaml":
        with open(grid_config_file, "r") as file:
            config = yaml.safe_load(file)
        config = squish_dict(config)
        for k, v in config.items():
            if type(v) != list:
                config[k] = [v]
        param_grid = ParameterGrid([config])
        configs = [unsquish_dict(param) for param in param_grid]

    elif config_ext == ".py":
        config_module = load_python_module(grid_config_file)
        if not hasattr(config_module, "configs"):
            raise KeyError(
                f"Config file at {grid_config_file} does not have an attribute named configs"
            )
        configs = config_module.configs
    else:
        raise ValueError(
            f"Config file at {grid_config_file} is not a valid YAML or python file"
        )

    total_num_params = len(configs)
    print(f"Config grid contains {total_num_params} combinations")

    # Load Slurm object from slurm_config_file
    slurm_module = load_python_module(slurm_config_file)
    if not hasattr(slurm_module, "slurm"):
        raise KeyError(
            f"slurm config file at {slurm_config_file} does not have an attribute named slurm"
        )
    slurm = slurm_module.slurm
    # slurm.add_arguments(wait=True)

    gid = configs[0].get("grid_id", None)
    same_gid = False
    if gid is not None:
        same_gid = all(conf.get("grid_id", None) == gid for conf in configs)

    gid_dir = None
    job_ids_file = None
    if same_gid:
        print(f"All configs have the same grid_id: {gid}")
        gid_dir = f"{file_root}/{GRID_OUTPUTS}/{gid}"
        os.makedirs(gid_dir, exist_ok=True)
        job_ids_file = os.path.join(gid_dir, "slurm_job_ids.txt")

    # make a temp dir to store the config files
    temp_configs_dir = os.path.join(gid_dir or file_root, TEMP_CONFIGS_DIR)
    os.makedirs(temp_configs_dir, exist_ok=True)

    jobs = []
    # Run the Sacred experiment with the provided adapter function and config
    for i, conf in enumerate(configs):
        print(
            "-" * 5
            + "GRID RUN INFO: "
            + f"Submitting run {i+1}/{total_num_params}"
            + "-" * 5
        )

        temp_config_file = os.path.join(temp_configs_dir, f"config_{i}.json")

        print(f"Saving the following config to {temp_config_file}")
        print(json.dumps(conf, indent=2))

        # Save config to a temp file
        with open(temp_config_file, "w") as file:
            json.dump(conf, file)

        slurm_command = (
            f"sorcerun run {python_file} {temp_config_file} --file_root {file_root} --auth_path {auth_path}"
            + ("-m" if mongo else "")
        )

        slurm.add_cmd(slurm_command)
        print(f"sbatch content:")
        print(slurm)

        cmd = "\n".join(
            (
                "sbatch" + " << EOF",
                slurm.script(shell="/bin/sh", convert=True),
                "EOF",
            )
        )
        slurm.run_cmds = slurm.run_cmds[:-1]

        # proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        # run the command and get the output
        out = subprocess.check_output(cmd, shell=True).decode("utf-8").strip()

        # extract slurm job id from out
        job_id = int(out.split()[-1])
        jobs.append(Job(job_id))

        # append the job id to the slurm_jobs.txt file if same_gid
        if same_gid:
            with open(job_ids_file, "a") as file:
                file.write(f"{job_id}\n")

        print(
            "-" * 5
            + "GRID RUN INFO: "
            + f"Finished submitting run {i+1}/{total_num_params}"
            + "-" * 5
        )

    print(f"Submitted {len(jobs)} jobs to slurm")

    if same_gid:
        print(f"Saved {len(jobs)} slurm job ids to {gid_dir}/slurm_job_ids.txt")

    time.sleep(10)
    poll_jobs(jobs)

    if post_process:
        # Post process grid and save xarray to netcdf
        # check if each config in configs has the same "grid_id" and assign it to gid
        if same_gid:
            print(f"Processing and saving grid to csv")
            process_and_save_grid_to_csv(gid, file_root=file_root)
            click.echo(f"Removing {job_ids_file}")
            os.remove(job_ids_file)
        else:
            print(
                f"Configs do not have the same grid_id."
                + " Skipping processing and saving grid to csv"
            )


@sorcerun.command()
@click.argument("grid_id", type=str)
@click.option(
    "--file_root",
    "-f",
    default=FILE_STORAGE_ROOT,
    type=click.Path(file_okay=False),
    help="Root directory for file storage",
)
def grid_to_netcdf(grid_id, file_root):
    save_dir = f"{file_root}/{GRID_OUTPUTS}/{grid_id}"
    # check if there is slurm_job_ids.txt in the grid_id directory
    job_ids_file = os.path.join(save_dir, "slurm_job_ids.txt")
    if os.path.exists(job_ids_file):
        click.echo(f"Slurm job ids found for grid with grid_id {grid_id}.")
        with open(job_ids_file, "r") as file:
            job_ids = file.read().strip().splitlines()
            jobs = [Job(int(job_id)) for job_id in job_ids]
            poll_jobs(jobs)

        # if we made it here, all jobs must have finished,
        # so remove slurm_job_ids.txt
        click.echo(f"Removing {job_ids_file}")
        os.remove(job_ids_file)

    click.echo(f"Processing and saving grid with grid_id {grid_id} to netcdf")
    process_and_save_grid_to_netcdf(grid_id, file_root=file_root)


@sorcerun.command()
@click.argument("grid_id", type=str)
@click.option(
    "--file_root",
    "-f",
    default=FILE_STORAGE_ROOT,
    type=click.Path(file_okay=False),
    help="Root directory for file storage",
)
def grid_to_csv(grid_id, file_root):
    save_dir = f"{file_root}/{GRID_OUTPUTS}/{grid_id}"
    # check if there is slurm_job_ids.txt in the grid_id directory
    job_ids_file = os.path.join(save_dir, "slurm_job_ids.txt")
    if os.path.exists(job_ids_file):
        click.echo(f"Slurm job ids found for grid with grid_id {grid_id}.")
        with open(job_ids_file, "r") as file:
            job_ids = file.read().strip().splitlines()
            jobs = [Job(int(job_id)) for job_id in job_ids]
            poll_jobs(jobs)

        # if we made it here, all jobs must have finished,
        # so remove slurm_job_ids.txt
        click.echo(f"Removing {job_ids_file}")
        os.remove(job_ids_file)

    click.echo(f"Processing and saving grid with grid_id {grid_id} to csv")
    process_and_save_grid_to_csv(grid_id, file_root=file_root)


@sorcerun.group()
def mongo():
    pass


@mongo.command()
def init():
    db_path = click.prompt(
        "Please enter the path where you'd like to store the MongoDB database files"
    )
    db_name = click.prompt("Enter the name of the database to store runs")
    username = click.prompt("Please enter a new username for the MongoDB database")
    password = click.prompt(
        "Please enter a new password for the MongoDB database", hide_input=True
    )

    click.echo("\n")

    init_mongodb(
        db_path=db_path,
        db_name=db_name,
        username=username,
        password=password,
    )

    click.echo(
        "Initialization complete. Use `sorcerun mongo start` to start the configured mongodb server."
    )


@mongo.command()
def start():
    with open(AUTH_FILE, "r") as f:
        authjson = json.loads(f.read())
        conf_path = authjson["conf_path"]

    print(conf_path)
    with mongodb_server(conf_path):
        while True:
            time.sleep(10)


@sorcerun.command()
def omniboard():
    """
    Run Omniboard using the details saved in {AUTH_FILE}.
    """
    try:
        with open(AUTH_FILE, "r") as f:
            auth_data = json.load(f)

        atlas = auth_data.get("atlas_connection_string", 0)
        mongodb_uri = (
            f"{atlas}/{auth_data['db_name']}"
            if atlas != 0
            else f"mongodb://{auth_data['client_kwargs']['username']}:{auth_data['client_kwargs']['password']}@{auth_data['client_kwargs']['host']}:{auth_data['client_kwargs']['port']}/{auth_data['db_name']}?authSource=admin"
        )
        # mongodb_uri = f'mongodb://{auth_data["client_kwargs"]["username"]}:{auth_data["client_kwargs"]["password"]}@127.0.0.1:27017/{auth_data["db_name"]}?authSource=admin'

        click.echo("Starting Omniboard...")
        with ExitStack() as stack:
            omniboard_process = subprocess.Popen(["omniboard", "--mu", mongodb_uri])
            stack.callback(omniboard_process.terminate)
            omniboard_process.wait()

    except (FileNotFoundError, json.JSONDecodeError):
        click.echo(
            f"Error: {AUTH_FILE} file not found or is not a valid JSON file. Run `sorcerun mongo init` to initialize this json file."
        )
    except subprocess.CalledProcessError as e:
        click.echo(f"Error occurred while running Omniboard: {e}")


@sorcerun.command()
def screen():
    screen_session_name = "sorcerun"

    # result = subprocess.Popen("screen -ls", shell=True, stdout=subprocess.PIPE)
    # output = "\n".join(str(a) for a in result.stdout.readlines())

    # click.echo(output)
    # if screen_session_name in output:
    #     click.echo(f"Screen session '{screen_session_name}' already exists.")
    #     return

    # Check if in conda environment
    conda_env = os.environ.get("CONDA_DEFAULT_ENV", None)
    activate_conda = f"conda activate {conda_env}\n" if conda_env else ""

    # Start new detached screen session
    screen_command = ["screen", "-dmS", screen_session_name]
    subprocess.Popen(screen_command)

    # # Start 'sorcerun mongo start' in first window
    # screen_command = [
    #     "screen",
    #     "-S",
    #     screen_session_name,
    #     "-X",
    #     "stuff",
    #     activate_conda + "sorcerun mongo start\n",
    # ]
    # subprocess.Popen(screen_command)

    # # Name the first window
    # screen_command = [
    #     "screen",
    #     "-S",
    #     screen_session_name,
    #     "-p",
    #     "0",
    #     "-X",
    #     "title",
    #     "mongo_start",
    # ]
    # subprocess.Popen(screen_command)

    # Create new window
    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-X",
        "screen",
        "-t",
        "mongodb",
    ]
    subprocess.Popen(screen_command)

    # Start 'sorcerun mongo start' in first window
    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-p",
        "mongodb",
        "-X",
        "stuff",
        activate_conda + "sorcerun mongo start\n",
    ]
    subprocess.Popen(screen_command)

    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-p",
        "1",
        "-X",
        "rename",
        "mongodb",
    ]
    subprocess.Popen(screen_command)

    # Create new window
    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-X",
        "screen",
        "-t",
        "omniboard",
    ]
    subprocess.Popen(screen_command)

    # Start 'sorcerun omniboard' in second window
    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-p",
        "omniboard",
        "-X",
        "stuff",
        activate_conda + "sorcerun omniboard\n",
    ]
    subprocess.Popen(screen_command)

    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-p",
        "2",
        "-X",
        "title",
        "omniboard",
    ]
    subprocess.Popen(screen_command)

    screen_command = [
        "screen",
        "-S",
        screen_session_name,
        "-p",
        "0",
        "-X",
        "kill",
    ]
    subprocess.Popen(screen_command)


# command to start grid_plotter streamlit app
@sorcerun.command()
def grid_plotter():
    try:
        this_file_path = os.path.abspath(__file__)
        this_dir = os.path.dirname(this_file_path)
        path_to_analysis_file = os.path.join(this_dir, "grid_plotter.py")

        with ExitStack() as stack:
            click.echo("Starting Streamlit app...")
            streamlit_process = subprocess.Popen(
                [
                    "streamlit",
                    "run",
                    path_to_analysis_file,
                ]
            )
            stack.callback(streamlit_process.terminate)
            streamlit_process.wait()
    except subprocess.CalledProcessError as e:
        click.echo(f"Error occurred while running streamlit: {e}")


if __name__ == "__main__":
    sorcerun()



# %%
from sorcerun.git_utils import get_tree_hash, get_commit_hash, get_repo, freeze_notebook
import sys

# %%
REPO = get_repo()
if REPO is None:
    raise Exception("Not in a git repository")
sys.path.append(f"{REPO.working_dir}/main")
# %% the following lines will be replaced when the notebook is frozen
COMMIT_HASH = get_commit_hash(REPO)
MAIN_TREE_HASH = get_tree_hash(REPO, "main")
# %%

# ... perform notebook things

# %% freeze the notebook
freeze_notebook(f"sorcerun/plot.py", REPO, COMMIT_HASH, MAIN_TREE_HASH)



AUTH_FILE = "sorcerun_auth.json"
MONGOD_HOST = "127.0.0.1"
MONGOD_PORT = 27017
GRID_OUTPUTS = "grid_outputs"
RUNS_DIR = "runs"
FILE_STORAGE_ROOT = "file_storage"
TEMP_CONFIGS_DIR = "temp_configs"
TEMPLATE_FILES = [
    "main/main.py",
    "main/config.py",
    "main/grid_config.py",
]
TIME_FORMAT = "%Y-%m-%d-%H-%M-%S"
